use std::collections::HashMap;

use anyhow::{bail, Result};

use crate::{handle::TextureHandle, Content};

pub(crate) mod syntax {
	use serde::Deserialize;
	// Raw serde syntax structs.
	// These are never seen by code outside this crate.

	#[derive(Debug, Deserialize)]
	pub struct Engine {
		pub thrust: f32,
		pub flare: Flare,
	}

	#[derive(Debug, Deserialize)]
	pub struct Flare {
		pub sprite_texture: String,
	}
}

/// Represents an (foward) engine outfit that may be attached to a ship.
#[derive(Debug, Clone)]
pub struct Engine {
	/// The name of this outfit
	pub name: String,

	/// How much thrust this engine produces
	pub thrust: f32,

	/// The flare sprite this engine creates.
	/// Its location and size is determined by a ship's
	/// engine points.
	pub flare_sprite_texture: TextureHandle,
}

impl crate::Build for Engine {
	type InputSyntax = HashMap<String, syntax::Engine>;

	fn build(engine: Self::InputSyntax, ct: &mut Content) -> Result<()> {
		for (engine_name, engine) in engine {
			let th = match ct.texture_index.get(&engine.flare.sprite_texture) {
				None => bail!(
					"In engine `{}`: texture `{}` doesn't exist",
					engine_name,
					engine.flare.sprite_texture
				),
				Some(t) => *t,
			};

			ct.engines.push(Self {
				name: engine_name,
				thrust: engine.thrust,
				flare_sprite_texture: th,
			});
		}

		return Ok(());
	}
}
