use cgmath::{Deg, InnerSpace};

use crate::{
	inputstatus::InputStatus,
	physics::{util, Physics, ShipHandle},
};

pub trait ShipBehavior
where
	Self: Send,
{
	fn update_controls(&mut self, physics: &mut Physics, input: &InputStatus, player: ShipHandle);
	fn get_handle(&self) -> ShipHandle;
}

pub struct Dummy {
	_handle: ShipHandle,
}

impl Dummy {
	pub fn new(handle: ShipHandle) -> Box<Self> {
		Box::new(Self { _handle: handle })
	}
}

impl ShipBehavior for Dummy {
	fn update_controls(
		&mut self,
		_physics: &mut Physics,
		_input: &InputStatus,
		_player: ShipHandle,
	) {
	}
	fn get_handle(&self) -> ShipHandle {
		return self._handle;
	}
}

pub struct Player {
	handle: ShipHandle,
}

impl Player {
	pub fn new(handle: ShipHandle) -> Box<Self> {
		Box::new(Self { handle })
	}
}

impl ShipBehavior for Player {
	fn update_controls(&mut self, physics: &mut Physics, input: &InputStatus, _player: ShipHandle) {
		let s = physics.get_ship_mut(&self.handle).unwrap();
		s.controls.left = input.key_left;
		s.controls.right = input.key_right;
		s.controls.guns = input.key_guns;
		s.controls.thrust = input.key_thrust;
	}

	fn get_handle(&self) -> ShipHandle {
		return self.handle;
	}
}

pub struct Point {
	handle: ShipHandle,
}

impl Point {
	pub fn new(handle: ShipHandle) -> Box<Self> {
		Box::new(Self { handle })
	}
}

impl ShipBehavior for Point {
	fn update_controls(&mut self, physics: &mut Physics, _input: &InputStatus, player: ShipHandle) {
		let (_, r) = physics.get_ship_body(&player).unwrap();
		let p = util::rigidbody_position(r);

		let (_, r) = physics.get_ship_body(&self.handle).unwrap();
		let t = util::rigidbody_position(r);
		let pa = util::rigidbody_rotation(r);
		let v = r.angvel();
		let d: Deg<f32> = (pa).angle(p - t).into();

		let s = physics.get_ship_mut(&self.handle).unwrap();
		s.controls.left = false;
		s.controls.right = false;

		if d < Deg(0.0) && v > -0.3 {
			s.controls.right = true;
		} else if d > Deg(0.0) && v < 0.3 {
			s.controls.left = true;
		}

		s.controls.guns = true;
		s.controls.thrust = false;
	}

	fn get_handle(&self) -> ShipHandle {
		return self.handle;
	}
}
